/**
 * 参数验证
 */
const {promisify} = require('util');
const matcher = require('micromatch');

const {sign, md5} = require('../helper/crypto')
const aes = require('../helper/aes')
const {isExclude} = require('../helper/excludes')
const {defaultKey} = require('../config')

const jsonBussError = function(res, message) {
	res.json({status: 300, body: null, message: message})
}

const jsonLoginError = function(res, message) {
	res.json({status: 1001, body: null, message: message})
}

const decodeToken = function(token) {
	if (!token) {
		return null
	}

	const tokenDecodeStr = aes.decode(token, defaultKey)
	const tokenDecodeArr = tokenDecodeStr.split(',')
	if (tokenDecodeArr.length !== 2) { // invalid token
		return null
	}
	return {
		userid: tokenDecodeArr[0],
		secret: tokenDecodeArr[1]
	}
}

const getPermission = function(authConfig, uri, method) {
	let result = undefined
	for (let item in authConfig) {
		if (authConfig.hasOwnProperty(item)) {
			if (matcher.isMatch(uri, item)) {
				const v = authConfig[item]
				result = v[method]
				break
			}
		}
	}
	return result
}


/**
 * @param config
 * 	urlPrefix: url 前缀，例如 /api
 * 	excludes: 排除的urls
 * 	isVerifyPermission: 是否需要验证permission
 * 	permissions: 设定的权限
 */
module.exports = function(config) {
	return async (req, res, next) => {

		const url = req.url
		if (isExclude(config.excludes, url)) {
			return next()
		}

		let tokenInfo
		const token = req.headers.v_token

		if (!config.isVerifyParameters) {
			tokenInfo = decodeToken(token)
			if (!tokenInfo) {
				return jsonBussError(res, 'invalid token')
			}
		} else {

			// const method = req.method.toLowerCase()
			const signature = req.headers.v_signature
			const timestamp = req.headers.v_timestamp

			if (!token) {
				return jsonBussError(res, 'lost token in headers')
			}
			if (!signature) {
				return jsonBussError(res, 'lost signature in headers')
			}
			if (!timestamp) {
				return jsonBussError(res, 'lost timestamp in headers')
			}

			tokenInfo = decodeToken(token)
			if (!tokenInfo) {
				return jsonBussError(res, 'invalid token')
			}

			const userid = tokenInfo.userid

			let pathname = req.pathname
			if (typeof pathname !== 'string') {
				pathname = req.path
			}

			const params = Object.assign({
				v_user: userid,
				v_timestamp: timestamp,
				v_url: pathname
			}, req.query, req.body)

			// verify signature
			const tSignature = sign(params, token)
			if (signature !== tSignature) {
				return jsonBussError(res, 'invalid signature')
			}
		}

		const userid = tokenInfo.userid
		const secret = tokenInfo.secret
		const getAsync = promisify(global.redisClient.get).bind(global.redisClient)
		let user = await getAsync(`account:${userid}`)
		if (!user) {
			return jsonLoginError(res, 'not login')
		}
		user = JSON.parse(user)
		// verify secret value
		// if (o.secret !== secret) {
		// 	return res.failure('invalid secret')
		// }

		// verify secret value
		// token如果被盗用，那么应尽快修改密码
		const hash = md5(user.id + user.password)
		if (hash !== secret) {
			return jsonLoginError(res, 'invalid secret')
		}
		if (!config.isVerifyPermission) { // not necessary verify permission
			res.locals.user = user
			return next()
		}

		const permissions = user.permissions
		if (!permissions || permissions.length === 0) {
			return jsonBussError(res, 'no permissions')
		}

		const method = req.method.toLowerCase()
		const urlPrefix = config.urlPrefix || ''
		const uri = url.substring(urlPrefix.length)

		// detect permission
		const permission = getPermission(config.permissions, uri, method)
		if (!permission) { // not found permission
			return jsonBussError(res, 'permission not defined')
		}
		if (!permissions.includes(permission)) {
			return jsonBussError(res, 'no permission')
		}

		// set user to res.locals
		res.locals.user = user

		return next()
	}
}